from .factor import Factor as Factor
from .factor import asfactor as asfactor
from .factor import gl as gl
from .format import round_scientific_notation as round_scientific_notation
from .format import truncate_name as truncate_name
from .linalg import cov2cor as cov2cor
from .lm import lm_fit as lm_fit
from .lm import lm_wfit as lm_wfit
from .logging import LOGGER as LOGGER
from .splines import ns as ns
from .splines import spline_design as spline_design
from .stats import dnbinom_mu as dnbinom_mu
from .stats import dnorm as dnorm
from .stats import pnorm as pnorm
from .stats import pt as pt
from .stats import rnbinom as rnbinom
