/* Generated by Cython 3.1.4 */

#ifndef __PYX_HAVE__inmoose__edgepy__edgepy_cpp
#define __PYX_HAVE__inmoose__edgepy__edgepy_cpp

#include "Python.h"

#ifndef __PYX_HAVE_API__inmoose__edgepy__edgepy_cpp

#ifdef CYTHON_EXTERN_C
    #undef __PYX_EXTERN_C
    #define __PYX_EXTERN_C CYTHON_EXTERN_C
#elif defined(__PYX_EXTERN_C)
    #ifdef _MSC_VER
    #pragma message ("Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.")
    #else
    #warning Please do not define the '__PYX_EXTERN_C' macro externally. Use 'CYTHON_EXTERN_C' instead.
    #endif
#else
    #define __PYX_EXTERN_C extern "C++"
#endif

#ifndef DL_IMPORT
  #define DL_IMPORT(_T) _T
#endif

__PYX_EXTERN_C PyArrayObject *vector2ndarray(std::vector<double>  const &);
__PYX_EXTERN_C double compute_unit_nb_deviance(double, double, double);
__PYX_EXTERN_C std::vector<double>  maximize_interpolant(std::vector<double> , __Pyx_memviewslice, int __pyx_skip_dispatch);

#endif /* !__PYX_HAVE_API__inmoose__edgepy__edgepy_cpp */

/* WARNING: the interface of the module init function changed in CPython 3.5. */
/* It now returns a PyModuleDef instance instead of a PyModule instance. */

/* WARNING: Use PyImport_AppendInittab("edgepy_cpp", PyInit_edgepy_cpp) instead of calling PyInit_edgepy_cpp directly from Python 3.5 */
PyMODINIT_FUNC PyInit_edgepy_cpp(void);

#if PY_VERSION_HEX >= 0x03050000 && (defined(__GNUC__) || defined(__clang__) || defined(_MSC_VER) || (defined(__cplusplus) && __cplusplus >= 201402L))
#if defined(__cplusplus) && __cplusplus >= 201402L
[[deprecated("Use PyImport_AppendInittab(\"edgepy_cpp\", PyInit_edgepy_cpp) instead of calling PyInit_edgepy_cpp directly.")]] inline
#elif defined(__GNUC__) || defined(__clang__)
__attribute__ ((__deprecated__("Use PyImport_AppendInittab(\"edgepy_cpp\", PyInit_edgepy_cpp) instead of calling PyInit_edgepy_cpp directly."), __unused__)) __inline__
#elif defined(_MSC_VER)
__declspec(deprecated("Use PyImport_AppendInittab(\"edgepy_cpp\", PyInit_edgepy_cpp) instead of calling PyInit_edgepy_cpp directly.")) __inline
#endif
static PyObject* __PYX_WARN_IF_PyInit_edgepy_cpp_INIT_CALLED(PyObject* res) {
  return res;
}
#define PyInit_edgepy_cpp() __PYX_WARN_IF_PyInit_edgepy_cpp_INIT_CALLED(PyInit_edgepy_cpp())
#endif

#endif /* !__PYX_HAVE__inmoose__edgepy__edgepy_cpp */
